import argparse
import subprocess
import shutil
import sys
import glob
import os
from functools import wraps
import threading

third_party_module_check_script = '/usr/libexec/upgrade-oc/check_user_installed_modules.sh'
system_service_check_script = '/usr/libexec/upgrade-oc/check_system_service.sh'
backup_script = '/usr/libexec/upgrade-oc/backup_restore.sh'

log_agent = '/usr/libexec/upgrade-oc/migration-log'

tmp_log_dir = '/var/log/upgrade-oc/'
leapp_log_dir = '/var/log/leapp/'

single_task_flag = os.path.join(tmp_log_dir, 'upgrade-oc-single-task.flag')
single_task_post_flag = os.path.join(tmp_log_dir, 'upgrade-oc-single-task-post.flag')

module_check_log = os.path.join(tmp_log_dir, 'module_check.log')
service_status_log = os.path.join(tmp_log_dir, 'service_status.txt')
summary_log = os.path.join(tmp_log_dir, 'summary.json')
backup_log = os.path.join(tmp_log_dir, 'backup.log')
postcheck_log = os.path.join(tmp_log_dir, 'postcheck.log')
user_app_check_log = os.path.join(tmp_log_dir, 'user_app_check.log')
leapp_preupgrade_log = 'leapp-preupgrade.log'
leapp_preupgrade_brief_log = os.path.join(tmp_log_dir, 'leapp-preupgrade-brief.log')
leapp_upgrade_log = os.path.join(leapp_log_dir, 'leapp-upgrade.log')
leapp_report_json = os.path.join(leapp_log_dir, 'leapp-report.json')

supported_versions = {
    'Tencent tlinux 2.2 (Final)': '2.2',
    'Tencent tlinux 2.6': '2.6',
    'TencentOS Server 4.2': '4.2',
    'TencentOS Server 4.0': '4.0',
    'TencentOS Server 3.2 (Final)': '3.2',
    'TencentOS Server 3.1 (Final)': '3.1',
}

def single_task_decorator(flag_file=single_task_flag):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # 检查标志文件是否存在
            if os.path.exists(flag_file):
                raise RuntimeError("Another task is already running.")
            
            # 创建标志文件
            with open(flag_file, 'w') as f:
                f.write('Task is running.')
            
            try:
                # 执行被装饰的函数
                result = func(*args, **kwargs)
            finally:
                # 删除标志文件
                if os.path.exists(flag_file):
                    os.remove(flag_file)

            return result
        return wrapper
    return decorator

def prepare_env(version):
    print("Executing prepare_env...")
    success = True
    try:
        download_leapp(supported_versions[version])
        download_rpm_check()

        if not os.path.exists(tmp_log_dir):
            os.makedirs(tmp_log_dir)

    except Exception as e:
        print("Failed to prepare environment: %s" % str(e))
        success = False
    
    print("Prepare_env completed successfully.")
    return success

def download_leapp(version):
    if version == "4.0" or version == "4.2":
        subprocess.call(["yum", "install", "-y", "leapp-upgrade-ts2tots4"])
    elif version == "3.1" or version == "3.2":
        subprocess.call(["yum", "install", "-y", "leapp-upgrade-ts2to3"])
    else:
        print("Unknown version: %s" % version)
        sys.exit(1)

def download_rpm_check():
    subprocess.call(["yum", "install", "-y", "rpm-check"])

def module_check(force, uuid):
    print("Executing module_check...")

    success = True

    try:
        # 执行 check_user_installed_modules.sh 脚本并将输出重定向到 module_check.log
        with open(module_check_log, 'w') as module_log:
            result = subprocess.call(third_party_module_check_script, stdout=module_log, stderr=subprocess.STDOUT)
            if result != 0:
                print("Module check failed.")
                if not force:
                    success = False
            else:
                print("Module check completed successfully.")
    except Exception as e:
        print("An error occurred while executing module_check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 module_check.log 文件是否存在
        if os.path.exists(module_check_log):
            # 上传 module_check.log 文件
            upload_log(module_check_log, uuid, "precheck")
        
        return success

def system_service_check(force, uuid):
    print("Executing system_service_check...")

    success = True

    try:
        # 执行 check_system_service.sh record 命令并将输出重定向到 service_status.txt
        result = subprocess.call([system_service_check_script, 'record', service_status_log], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if result != 0:
            print("Service status check failed.")
            if not force:
                success = False
        else:
            print("Service status check completed successfully.")
    except Exception as e:
        print("An error occurred while executing service status check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 service_status.txt 文件是否存在
        if os.path.exists(service_status_log):
            # 上传 service_status.txt 文件
            upload_log(service_status_log, uuid, "precheck")
        return success

def get_pretty_name():
    file_path='/etc/os-release'

    try:
        with open(file_path, 'r') as file:
            for line in file:
                if line.startswith('PRETTY_NAME='):
                    # 去掉前缀 'PRETTY_NAME=' 并去掉引号
                    pretty_name = line[len('PRETTY_NAME='):].strip().strip('"')
                    return pretty_name
    except FileNotFoundError:
        print("File {} not found.".format(file_path))
    except Exception as e:
        print("An error occurred: {}".format(str(e)))

    return None

def rpm_check(userapp, serverip, target, force, uuid):
    print("Executing rpm_check...")

    success = True

    try:
        # 执行 rpm-check 命令
        if userapp and serverip:
            # 4.2 4.0, we all consider them as 4, 2.2 2.6 as 2
            current_ver = supported_versions[get_pretty_name()].split('.')[0]
            target_version = supported_versions[target].split('.')[0]
            serverip = serverip + '/ts' + current_ver + 'tots' + target_version

            rpm_check_cmd = ['rpm-check', '--checkabi', '--check-local', userapp, '--search-remote', serverip, '--with-symbol']
            result = subprocess.call(rpm_check_cmd)
            if result != 0:
                print("RPM check failed.")
                if not force:
                    success = False
            else:
                print("RPM check completed successfully.")
        else:
            print("Skipping RPM check due to missing userapp or serverip.")
    except Exception as e:
        print("An error occurred while executing rpm-check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 summary.log 文件是否存在
        if os.path.exists("./summary.json"):
            # 上传 summary.log 文件
            shutil.move("./summary.json", summary_log)
            upload_log(summary_log, uuid, "precheck")
        return success

def leapp_check(backup_dir, exclude_dirs, force, uuid):
    print("Executing leapp_check...")

    success = True

    try:
        # 执行 leapp --preupgrade 命令
        with open(leapp_preupgrade_brief_log, 'w') as preupgrade_log:
            result = subprocess.call(['leapp', 'preupgrade', '--tencentos'], stdout=preupgrade_log, stderr=preupgrade_log)
            if result != 0:
                print("Leapp preupgrade failed.")
                success = False
            else:
                print("Leapp preupgrade completed successfully.")

        ret = backup("check", backup_dir, exclude_dirs, uuid)
        if not force and not ret:
            success = False
    except Exception as e:
        print("An error occurred while executing leapp-check: %s" % str(e))
        if not force:
            success = False
    finally:
        # 检查 leapp-preupgrade.log 文件是否存在
        if os.path.exists(leapp_preupgrade_log):
            # 上传 leapp-preupgrade.log 文件
            upload_log(leapp_preupgrade_log, uuid, "precheck")
        
        if os.path.exists(leapp_report_json):
            upload_log(leapp_report_json, uuid, "precheck")

        return success

@single_task_decorator()
def precheck(backup_dir, exclude_dirs, userapp=None, serverip=None, target=None, force=False, uuid=None):
    print("Executing precheck...")

    any_failure = False

    try:
        if not module_check(force, uuid):
            any_failure = True

        if not system_service_check(force, uuid):
            any_failure = True

        if not rpm_check(userapp, serverip, target, force, uuid):
            any_failure = True

        if not leapp_check(backup_dir, exclude_dirs, force, uuid):
            any_failure = True

    except Exception as e:
        print("An unexpected error occurred during precheck: {}".format(e))
        any_failure = True

    finally:
        if any_failure:
            print("Precheck completed with failures.")
            return False

        print("Precheck completed successfully.")
        return True

def backup(action, backup_dir, exclude_dirs, uuid):
    print("Executing backup...")

    success = True

    try:
        # 构建 backup_restore.sh 命令
        exclude_dirs_str = ','.join(exclude_dirs)
        backup_cmd = [
            backup_script,
            '-a', action,
            '-d', backup_dir,
            '-e', exclude_dirs_str,
        ]

        # 执行 backup_restore.sh 命令
        with open(backup_log, 'a') as backuplog:
            ret = subprocess.call(backup_cmd, stdout=backuplog, stderr=backuplog)
            if ret != 0:
                print("Backup failed with return code {}".format(ret))
                success = False
            else:
                print("Backup completed successfully.")

    except subprocess.CalledProcessError as e:
        print("An error occurred during backup: {}".format(e))
        success = False
    finally:
        # 检查 backup.log是否存在
        if action == "backup" and os.path.exists(backup_log):
            upload_log(backup_log, uuid, "upgrade")

        return success

def log_agent_finish(uuid, stage):
    command = 'echo "" | {} pipe-stdin --stage={} --status=finish --uuid={}'.format(log_agent, stage, uuid)
    result = subprocess.call(command, shell=True)

@single_task_decorator()
def upgrade(backup_dir, exclude_dirs, version, uuid):
    print("Executing upgrade...")

    success = True

    # 先执行备份操作
    if not backup("backup", backup_dir, exclude_dirs, uuid):
        print("Backup failed. Upgrade aborted.")
        success = False
        return success

    try:
        # 构建 leapp upgrade 命令
        command = ['leapp', 'upgrade', '--tencentos']
        if version:
            command.extend(['--target', supported_versions[version]])
        
        # 执行 leapp upgrade 命令
        if uuid:
            # 构建 migration-log 命令
            log_command = [log_agent, 'log-file', '--filepath={}'.format(leapp_upgrade_log), '--uuid={}'.format(uuid), '--status=begin', '--stage=upgrade']

            with subprocess.Popen(command, stdout=subprocess.PIPE) as proc1:
                with subprocess.Popen(log_command, stdin=proc1.stdout, stdout=subprocess.PIPE) as proc2:
                    proc1.stdout.close()  # Allow proc1 to receive a SIGPIPE if proc2 exits.
                    output, errors = proc2.communicate()
        else:
            subprocess.call(command)

        print("Upgrade completed successfully.")
    except subprocess.CalledProcessError as e:
        print("An error occurred during upgrade: {}".format(e))
        success = False
    finally:
        ## finish log report
        if uuid:
            log_agent_finish(leapp_upgrade_log, uuid, "upgrade")

        if os.path.exists(leapp_report_json):
            upload_log(leapp_report_json, uuid, "upgrade")

        # 如果成功，重启进入新系统
        if success:
            subprocess.call("reboot")

        return success

@single_task_decorator(single_task_post_flag)
def postcheck(commands=None, uuid=None):
    print("Executing postcheck...")

    try:
        # 执行用户提供的命令
        if commands:
            for command in commands.split(','):
                print("Running command: {}".format(command))
                with open(user_app_check_log, 'a') as user_log:
                    result = subprocess.call(command, stdout=user_log, stderr=user_log, shell=True)
                if result != 0:
                    print("Command '{}' failed with return code {}.".format(command, result))
                    return False

        # 执行 check_system_service.sh compare 命令
        print("Running system service comparison check...")
        with open(postcheck_log, 'w') as service_log:
            subprocess.call([system_service_check_script, 'compare', service_status_log], stdout=service_log, stderr=service_log)
        print("System service comparison check completed successfully.")
    except subprocess.CalledProcessError as e:
        print("An error occurred during system service comparison check: {}".format(e))
        return False
    finally:
        if os.path.exists(user_app_check_log):
            upload_log(user_app_check_log, uuid, "postcheck")
        
        if os.path.exists(postcheck_log):
            upload_log(postcheck_log, uuid, "postcheck")

    print("Postcheck completed successfully.")
    return True

@single_task_decorator(single_task_post_flag)
def clean():
    print("Executing clean...")

    if os.path.exists(tmp_log_dir):
        shutil.rmtree(tmp_log_dir)
        print("Directory {} has been deleted.".format(tmp_log_dir))
    else:
        print("Directory {} does not exist.".format(tmp_log_dir))

    if os.path.exists(tmp_log_dir):
        shutil.rmtree(tmp_log_dir)
        print("Directory {} has been deleted.".format(tmp_log_dir))
    else:
        print("Directory {} does not exist.".format(tmp_log_dir))

    # 卸载软件包
    packages = [
        'rpm-check',
        'rpm-check-tools',
        'leapp'
    ]

    for package in packages:
        try:
            print("Uninstalling package: {}".format(package))
            subprocess.call(['yum', 'remove', '-y', package])
            print("Package {} uninstalled successfully.".format(package))
        except subprocess.CalledProcessError as e:
            print("An error occurred while uninstalling package {}: {}".format(package, e))
            return False

    print("Clean completed successfully.")
    return True


def create_systemd_service(userscript, backup_dir, uuid, target):

    service_name = "upgrade_oc_postcheck"

    if supported_versions[target].startswith("4."):
        version_dir = "ts2to4"
    else:
        version_dir = "ts3to3"

    leapp_bundle_dir = "/usr/share/leapp-repository/repositories/system_upgrade/{}/files/bundled-rpms/".format(version_dir)
    upgrade_oc_tool_dir = "/usr/libexec/upgrade-oc/"

    if not copy_new_version_to_leapp_dir(upgrade_oc_tool_dir, leapp_bundle_dir, "upgrade-oc*.rpm"):
        print("Failed to copy new version of upgrade-oc to leapp directory.")
        return False

    # 定义服务单元文件内容
    service_content = """
[Unit]
Description=Custom Upgrade Service
After=network.target

[Service]
Type=simple
ExecStart=upgrade-oc --postcheck --userscript '{}' --backup-dir {} --uuid {}
Restart=on-failure

[Install]
WantedBy=multi-user.target
""".format(userscript, backup_dir, uuid)

    # 服务单元文件路径
    service_file_path = "/etc/systemd/system/{}.service".format(service_name)

    try:
        # 将服务单元文件内容写入文件
        with open(service_file_path, 'w') as service_file:
            service_file.write(service_content)

        # 重新加载 systemd 配置
        subprocess.call(['systemctl', 'daemon-reload'])

        # 启用服务，使其开机自启动
        subprocess.call(['systemctl', 'enable', service_name])
    except Exception as e:
        print("Failed to create service {}: {}".format(service_name, str(e)))
        return False
    finally:
        print("Service {} created and enabled successfully.".format(service_name))
        return True

def copy_new_version_to_leapp_dir(source_dir, destination_dir, pattern):
    # 确保目标目录存在
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)

    # 查找源目录中的 RPM 文件
    rpm_files = glob.glob(os.path.join(source_dir, pattern))

    if not rpm_files:
        print("No RPM files found in the source directory.")
        return False

    # 复制每个 RPM 文件到目标目录
    for rpm_file in rpm_files:
        shutil.copy(rpm_file, destination_dir)
        print("Copied {} to {}".format(rpm_file, destination_dir))
    
    return True

def clean_backup(backup_dir):
    # 删除备份
    if os.path.exists(backup_dir):
        try:
            shutil.rmtree(backup_dir)
            print("Deleted backup directory: {}".format(backup_dir))
        except Exception as e:
            print("An error occurred while deleting backup directory {}: {}".format(backup_dir, e))
            
            return False
    
    print("Backup deleted successfully.")
    return True

def upload_logs_dir(uuid):

    if not uuid:
        return True

    res = True

    ret = subprocess.call([log_agent, 'push-dir', '--dirpath={}'.format(leapp_log_dir), '--uuid={}'.format(uuid)])
    if ret != 0:
        print("Failed to upload logs directory {}: {}".format(leapp_log_dir, ret))
        res = False
    else:
        print("Uploaded logs directory {} successfully.".format(leapp_log_dir))

    ret = subprocess.call([log_agent, 'push-dir', '--dirpath={}'.format(tmp_log_dir), '--uuid={}'.format(uuid)])
    if ret != 0:
        print("Failed to upload logs directory {}: {}".format(tmp_log_dir, ret.stderr))
        res = False
    else:
        print("Uploaded logs directory {} successfully.".format(tmp_log_dir))

    return res

def upload_log(log_path, uuid, stage):

    if not uuid:
        return True

    res = True

    ret = subprocess.call([log_agent, 'push-file', '--filepath={}'.format(log_path), '--uuid={}'.format(uuid), '--stage={}'.format(stage)])
    if ret != 0:
        print("Failed to upload log file {}: {}".format(ret.stderr))
        
        res = False

    return res

def main():
    parser = argparse.ArgumentParser(description="OC Upgrade Tool")
    parser.add_argument('--precheck', action='store_true', help='Execute precheck process')
    parser.add_argument('--upgrade', action='store_true', help='Execute upgrade process')
    parser.add_argument('--auto', action='store_true', help='Automatically execute upgrade/postcheck/clean if precheck succeeds')
    parser.add_argument('--postcheck', action='store_true', help='Execute postcheck process')
    parser.add_argument('--userapp', type=str, help='User application for precheck')
    parser.add_argument('--serverip', type=str, help='Server IP for precheck')
    parser.add_argument('--userscript', type=str, help='User script for postcheck, support comma-separated commands and entire script path', metavar='[script or CMD list]')
    parser.add_argument('--backup-dir', type=str, help='Backup directory for upgrade')
    parser.add_argument('--exclude-dirs', nargs='+', help='Directories to exclude from upgrade')
    parser.add_argument('--force', action='store_true', help='Force upgrade even if precheck fails(exclude leapp preupgrade)')
    parser.add_argument('--clean', action='store_true', help='Execute clean process')
    parser.add_argument('--uuid', type=str, help='Upload logs to Tmanager with UUID', metavar='UUID')
    parser.add_argument('--target', type=str, help='Target version for upgrade')
    args = parser.parse_args()

    exit_code = 0

    if args.target and args.target not in supported_versions:
        exit_code = 1
        parser.error("Unsupported target version: {}".format(args.target))

    elif args.precheck:
        if not args.userapp or not args.serverip or not args.backup_dir or not args.exclude_dirs or not args.target:
            exit_code = 1
            parser.error("Missing required arguments for precheck")
        
        if not prepare_env(args.target):
            print("Failed to prepare environment.")
            exit_code = 1
            sys.exit(exit_code)

        if not precheck(args.backup_dir, args.exclude_dirs, args.userapp, args.serverip, args.target, args.force, args.uuid):
            print("Precheck failed.")
            exit_code = 1
            sys.exit(exit_code)

        if args.auto:
            if not create_systemd_service(args.userscript, args.backup_dir, args.uuid, args.target):
                print("Failed to create systemd service.")
                exit_code = 1

            if not upgrade(args.backup_dir, args.exclude_dirs, args.target, args.uuid):
                print("Upgrade failed.")
                exit_code = 1

    elif args.upgrade:
        if not args.backup_dir or not args.exclude_dirs:
            parser.error("Missing required arguments for upgrade")

        if not upgrade(args.backup_dir, args.exclude_dirs, args.target, args.uuid):
            print("Upgrade failed.")
            exit_code = 1

    elif args.postcheck:
        if not args.userscript or not args.backup_dir:
            parser.error("Missing required arguments for postcheck")

        ret = postcheck(args.userscript, args.uuid)

        if not ret:
            print("Postcheck failed.")
            exit_code = 1
        
        if upload_logs_dir(args.uuid):
            # 上传日志失败，不清理日志
            clean()

        clean_backup(args.backup_dir)

    elif args.clean:
        if not args.backup_dir:
            parser.error("Missing required arguments for clean")

        clean()
        clean_backup(args.backup_dir)
    
    sys.exit(exit_code)

if __name__ == "__main__":
    main()
